"""Run shell commands.

This module provides various methods to run shell commands, on both local and remote hosts, as current or another user.
There many variant of the methods to allow fine-gain control of parameter passing and error handling.

4 different implementations are provided:

1. LocalShell allows to run command on local host as various users. It is the most feature-rich one, allowing
   interactive I/O from/to the terminal.
2. SSHShell allows to run command on both local and remote hosts. When running on a remote host, it creates a direct
   connection from a specified local_user to the destination host and user. User input from terminal is not allowed, as
   the command is passed through the stdin.
3. ClusterShell runs command on cluster nodes. It leverages su, sudo and ssh to obtain a appreciated session on a
   destination node. It is only available after ssh bootstrap as it depends on the knowledge about cluster node and user
   configurations.
4. ShellUtils runs command on local host as current user. It is a simple wrapper around subprocess module.

The LocalShell and SSHShell is expected to be used in ssh bootstrap. Once the ssh bootstrap finishes, ClusterShell should
be used.
"""
import logging
import os
import pwd
import re
import socket
import subprocess
import typing
from io import StringIO

from . import constants
from .pyshim import cache
from . import user_of_host
from .user_of_host import UserOfHost

import crmsh.options

logger = logging.getLogger(__name__)


class Error(ValueError):
    def __init__(self, msg, cmd):
        super().__init__(msg)
        self.cmd = cmd


class AuthorizationError(Error):
    def __init__(self, cmd: str, host: typing.Optional[str], user: str, msg: str):
        super().__init__(
            'Failed to run command {cmd} on {optional_user}{host}: {msg} {diagnose}'.format(
                optional_user=f'{user}@' if user is not None else '',
                host=host, msg=msg, cmd=cmd,
                diagnose=self.diagnose(),
            ),
            cmd
        )
        self.host = host
        self.user = user

    def diagnose(self) -> str:
        return ''


class NonInteractiveSSHAuthorizationError(AuthorizationError):

    def diagnose(self) -> str:
        ret = super().diagnose()
        if not ret:
            return 'Please configure passwordless authentication with "crm cluster init ssh" and "crm cluster join ssh"'
        return ret


class CommandFailure(Error):
    def __init__(self, cmd: str, host: typing.Optional[str], user: typing.Optional[str], msg: str):
        if host is None and user is None:
            super().__init__("Failed to run '{}': {}".format(cmd, msg), cmd)
        elif user is None:
            super().__init__("Failed to run command on {}: '{}': {}".format(host, cmd, msg), cmd)
        elif host is None:
            super().__init__("Failed to run command as {}: '{}': {}".format(user, cmd, msg), cmd)
        else:
            super().__init__("Failed to run command as {}@{}: '{}': {}".format(user, host, cmd, msg), cmd)
        self.host = host
        self.user = user
        self.cmd = cmd
        self.msg = msg

    def __reduce__(self):
        '''
        Return a tuple that Python will use for pickling this object
        '''
        return CommandFailure, (self.cmd, self.host, self.user, self.msg)


class Utils:
    @staticmethod
    def decode_str(x: bytes):
        try:
            return x.decode('utf-8')
        except UnicodeDecodeError as e:
            logger.debug('UTF-8 decode failure', exc_info=e)
            return x.decode('utf-8', errors='backslashreplace')


class LocalShell:
    """Provides methods to run commands on localhost, both as current user and switching to another user"""
    @staticmethod
    @cache
    def hostname():
        return socket.gethostname()

    @staticmethod
    @cache
    def geteuid() -> int:
        return os.geteuid()

    @staticmethod
    @cache
    def get_effective_user_name() -> str:
        return pwd.getpwuid(LocalShell.geteuid()).pw_name

    def __init__(self, additional_environ: typing.Dict[str, str] = None):
        self.additional_environ = additional_environ
        self.preserve_env = additional_environ.keys() if additional_environ is not None else None

    def can_run_as(self, user: str):
        return self.geteuid() == 0 or self.get_effective_user_name() == user

    def su_subprocess_run(
            self,
            user: typing.Optional[str],
            cmd: str,
            tty=False,
            **kwargs,
    ):
        """Call subprocess.run as another user.

        This variant is the most flexible one as it pass unknown kwargs to the underlay subprocess.run. However, it
        accepts only cmdline but not argv, as the argv is used internally to switch user.
        """
        if user is None or self.get_effective_user_name() == user:
            args = ['/bin/sh', '-c', cmd]
        elif 0 == self.geteuid():
            args = ['su', user, '--login', '-s', '/bin/sh', '-c', cmd]
            if tty:
                args.append('--pty')
            if self.preserve_env:
                args.append('-w')
                args.append(','.join(self.preserve_env))
        else:
            raise AuthorizationError(
                cmd, None, user,
                f"non-root user '{self.get_effective_user_name()}' cannot switch to another user"
            )
        if not self.additional_environ:
            logger.debug('su_subprocess_run: %s, %s', args, kwargs)
            env = os.environ    # bsc#1205925
        else:
            logger.debug('su_subprocess_run: %s, env=%s, %s', args, self.additional_environ, kwargs)
            env = dict(os.environ)
            env.update(self.additional_environ)
        return subprocess.run(args, env=env, **kwargs)

    def get_rc_stdout_stderr_raw(self, user: typing.Optional[str], cmd: str, input: typing.Optional[bytes] = None):
        result = self.su_subprocess_run(
            user, cmd,
            input=input,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode, result.stdout, result.stderr

    def get_rc_stdout_stderr(self, user: typing.Optional[str], cmd: str, input: typing.Optional[str] = None):
        rc, stdout, stderr = self.get_rc_stdout_stderr_raw(user, cmd, input.encode('utf-8') if input is not None else None)
        return rc, Utils.decode_str(stdout).strip(), Utils.decode_str(stderr).strip()

    def get_rc_and_error(
            self,
            user: typing.Optional[str],
            cmd: str,
    ) -> typing.Tuple[int, typing.Optional[str]]:
        """Run a command for its side effects. Returns (rc, error_message)

        If the return code is 0, outputs from the command will be ignored and (0, None) is returned.
        If the return code is not 0, outputs from the stdout and stderr is combined as a single message.
        """
        result = self.su_subprocess_run(
            user, cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            start_new_session=True,
        )
        if result.returncode == 0:
            return 0, None
        else:
            return result.returncode, Utils.decode_str(result.stdout).strip()

    def get_stdout_or_raise_error(
            self,
            user: typing.Optional[str],
            cmd: str,
            success_exit_status: typing.Optional[typing.Set[int]] = None,
    ) -> str:
        result = self.su_subprocess_run(
            user, cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        to_raise = False
        if success_exit_status is None:
            if result.returncode != 0:
                to_raise = True
        else:
            if result.returncode not in success_exit_status:
                to_raise = True
        if not to_raise:
            return Utils.decode_str(result.stdout).strip()
        else:
            raise CommandFailure(cmd, None, user, Utils.decode_str(result.stderr).strip())


class SSHShell:
    """Provides methods to run commands on both local and remote hosts as various users.

    For remote commands, SSH sessions are created to the destination host and user from a specified local_user.
    """
    def __init__(self, local_shell: LocalShell, local_user):
        self.local_shell = local_shell
        self.local_user = local_user

    def can_run_as(self, host: typing.Optional[str], user: str) -> bool:
        # This method does not call subprocess_run_without_input. The reason may be some of the callers expect that ssh
        # is used even if the destination host is localhost.
        if host is None or host == self.local_shell.hostname():
            return self.local_shell.can_run_as(user)
        else:
            ssh_options = "-o StrictHostKeyChecking=no -o EscapeChar=none -o ConnectTimeout=15"
            ssh_cmd = "ssh {} -T -o Batchmode=yes {}@{} true".format(ssh_options, user, host)
            rc, output = self.local_shell.get_rc_and_error(self.local_user, ssh_cmd)
            return rc == 0

    def get_rc_and_error(
            self,
            host: typing.Optional[str],
            user: str,
            cmd: str,
    ) -> typing.Tuple[int, typing.Optional[str]]:
        """Run a command for its side effects. Returns (rc, error_message)

        If the return code is 0, outputs from the command will be ignored and (0, None) is returned.
        If the return code is not 0, outputs from the stdout and stderr is combined as a single message.
        """
        result = self.subprocess_run_without_input(
            host, user, cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            start_new_session=True,
        )
        if result.returncode == 0:
            return 0, None
        else:
            return result.returncode, Utils.decode_str(result.stdout).strip()

    def subprocess_run_without_input(self, host: typing.Optional[str], user: str, cmd: str, **kwargs):
        assert 'input' not in kwargs and 'stdin' not in kwargs
        if host is None or host == self.local_shell.hostname():
            if user == self.local_shell.get_effective_user_name():
                args = ['/bin/sh']
            else:
                args = ['sudo', '-H', '-u', user, '/bin/sh']
            return subprocess.run(
                args,
                input=cmd.encode('utf-8'),
                env=os.environ,     # bsc#1205925
                **kwargs,
            )
        else:
            return self.local_shell.su_subprocess_run(
                self.local_user,
                'ssh {} {}@{} /bin/sh'.format(constants.SSH_OPTION, user, host),
                input=cmd.encode('utf-8'),
                **kwargs,
            )


class ClusterShell:
    """Provides methods to run commands on both local and remote cluster nodes.

    For remote nodes, the local and remote user used for SSH sessions are determined from cluster configuration recorded
    during bootstrap.
    """
    def __init__(
            self,
            local_shell: LocalShell,
            user_of_host: UserOfHost,
            forward_ssh_agent: bool = False,
            raise_ssh_error: bool = False,  # whether to raise AuthorizationError when ssh returns with 255
    ):
        self.local_shell = local_shell
        self.user_of_host = user_of_host
        self.forward_ssh_agent = forward_ssh_agent
        self.raise_ssh_error = raise_ssh_error

    def can_run_as(self, host: typing.Optional[str], user: str) -> bool:
        try:
            result = self.subprocess_run_without_input(
                host, user, 'true',
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except crmsh.sh.AuthorizationError:
            return False
        except user_of_host.UserNotFoundError:
            return False
        return 0 == result.returncode

    def subprocess_run_without_input(self, host: typing.Optional[str], user: typing.Optional[str], cmd: str, **kwargs):
        assert 'input' not in kwargs and 'stdin' not in kwargs
        if host is None or host == self.local_shell.hostname():
            if user is None:
                return subprocess.run(
                    ['/bin/sh'],
                    input=cmd.encode('utf-8'),
                    env=os.environ,  # bsc#1205925
                    **kwargs,
                )
            else:
                return self.local_shell.su_subprocess_run(
                    user, cmd,
                    **kwargs,
                )
        else:
            if user is None:
                user = 'root'
            local_user, remote_user = self.user_of_host.user_pair_for_ssh(host)
            result = self.local_shell.su_subprocess_run(
                local_user,
                'ssh {} {} -o BatchMode=yes {}@{} sudo -H -u {} {} /bin/sh'.format(
                    '-A' if self.forward_ssh_agent else '',
                    constants.SSH_OPTION,
                    remote_user,
                    host,
                    user,
                    '--preserve-env=SSH_AUTH_SOCK' if self.forward_ssh_agent else '',
                ),
                input=cmd.encode('utf-8'),
                start_new_session=True,
                **kwargs,
            )
            if self.raise_ssh_error and result.returncode == 255:
                raise NonInteractiveSSHAuthorizationError(
                    cmd, host, remote_user,
                    Utils.decode_str(result.stderr).strip() if result.stderr is not None else ''
                )
            else:
                return result

    def get_rc_and_error(self, host: typing.Optional[str], user: str, cmd: str):
        result = self.subprocess_run_without_input(
            host, user, cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if result.returncode == 0:
            return 0, None
        else:
            return result.returncode, Utils.decode_str(result.stdout).strip()

    def get_rc_stdout_stderr_raw_without_input(self, host, cmd) -> typing.Tuple[int, bytes, bytes]:
        result = self.subprocess_run_without_input(
            host, None, cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.returncode, result.stdout, result.stderr

    def get_rc_stdout_stderr_without_input(self, host, cmd) -> typing.Tuple[int, str, str]:
        rc, stdout, stderr = self.get_rc_stdout_stderr_raw_without_input(host, cmd)
        return rc, Utils.decode_str(stdout).strip(), Utils.decode_str(stderr).strip()

    def get_rc_output_without_input(self, host, cmd) -> typing.Tuple[int, str]:
        result = self.subprocess_run_without_input(
            host, None, cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )
        return result.returncode, Utils.decode_str(result.stdout).strip()

    def get_stdout_or_raise_error(
            self,
            cmd: str,
            host: typing.Optional[str] = None,
            success_exit_status: typing.Optional[typing.Set[int]] = None,
    ):
        rc, stdout, stderr = self.get_rc_stdout_stderr_raw_without_input(host, cmd)
        to_raise = False
        if success_exit_status is None:
            if rc != 0:
                to_raise = True
        else:
            if rc not in success_exit_status:
                to_raise = True
        if not to_raise:
            return Utils.decode_str(stdout).strip()
        else:
            raise CommandFailure(cmd, host, None, Utils.decode_str(stderr).strip())

    def ssh_to_localhost(self, user: typing.Optional[str], cmd: str, **kwargs):
        if user is None:
            user = 'root'
        host = self.local_shell.hostname()
        local_user, remote_user = self.user_of_host.user_pair_for_ssh(host)
        result = self.local_shell.su_subprocess_run(
            local_user,
            'ssh {} {} {}@{} sudo -H -u {} {} /bin/sh'.format(
                '-A' if self.forward_ssh_agent else '',
                constants.SSH_OPTION,
                remote_user,
                host,
                user,
                '--preserve-env=SSH_AUTH_SOCK' if self.forward_ssh_agent else '',
            ),
            input=cmd.encode('utf-8'),
            **kwargs,
        )
        if self.raise_ssh_error and result.returncode == 255:
            raise AuthorizationError(cmd, host, remote_user, Utils.decode_str(result.stderr).strip())
        else:
            return result


class ShellUtils:
    CONTROL_CHARACTER_PATTER = re.compile('[\u0000-\u001F]')

    @classmethod
    def get_stdout(cls, cmd, input_s=None, stderr_on=True, shell=True, raw=False):
        '''
        Run a cmd, return stdout output.
        Optional input string "input_s".
        stderr_on controls whether to show output which comes on stderr.
        '''
        if crmsh.options.regression_tests:
            print(".EXT", cmd)
        proc = subprocess.Popen(
            cmd,
            shell=shell,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL if stderr_on else subprocess.PIPE,
            env=os.environ,  # bsc#1205925
        )
        stdout_data, _ = proc.communicate(input_s)
        if raw:
            return proc.returncode, stdout_data
        else:
            if isinstance(stdout_data, bytes):
                stdout_data = Utils.decode_str(stdout_data)
        return proc.returncode, stdout_data.strip()

    @classmethod
    def get_stdout_stderr(cls, cmd, input_s=None, shell=True, raw=False, no_reg=False, timeout=None, mix_stderr=False):
        '''
        Run a cmd, return (rc, stdout, stderr)
        Parameters:
            mix_stderr: whether to mix stderr data into stdout
        '''
        if crmsh.options.regression_tests and not no_reg:
            print(".EXT", cmd)
        proc = subprocess.Popen(
            cmd,
            shell=shell,
            stdin=input_s and subprocess.PIPE or None,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if mix_stderr else subprocess.PIPE,
            env=os.environ,  # bsc#1205925
        )
        # will raise subprocess.TimeoutExpired if set timeout
        stdout_data, stderr_data = proc.communicate(input_s, timeout=timeout)
        if raw:
            return proc.returncode, stdout_data, stderr_data
        else:
            if isinstance(stdout_data, bytes):
                stdout_data = Utils.decode_str(stdout_data).strip()
            if isinstance(stderr_data, bytes):
                stderr_data = Utils.decode_str(stderr_data).strip()
        return proc.returncode, stdout_data, stderr_data


class ClusterShellAdaptorForLocalShell(ClusterShell):
    """A adaptor to wrap a LocalShell as a ClusterShell.

    Some modules depend on shell and are called both during bootstrap and after bootstrap. Use a LocalShell as their
    implementation in bootstrap make the difference more explicit, avoid dependency on outdated cluster configurations
    (for example, the configurations left from previous cluster bootstrap) and help to catch errors in tests.
    """
    def __init__(self, local_shell: LocalShell):
        super().__init__(local_shell, None)

    def subprocess_run_without_input(self, host: str, user: typing.Optional[str], cmd: str, **kwargs):
        assert 'input' not in kwargs and 'stdin' not in kwargs
        assert host is None or host == self.local_shell.hostname()
        if user is None:
            user = 'root'
        return self.local_shell.su_subprocess_run(user, cmd, **kwargs)


def cluster_shell():
    return ClusterShell(
        LocalShell(additional_environ={'SSH_AUTH_SOCK': os.environ.get('SSH_AUTH_SOCK', '')}),
        user_of_host.instance(),
        forward_ssh_agent=True,
        raise_ssh_error=True,
    )




# -*- coding: utf-8 -*-
"""
    pygments.lexers.dsls
    ~~~~~~~~~~~~~~~~~~~~

    Lexers for various domain-specific languages.

    :copyright: Copyright 2006-2015 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""
from __future__ import unicode_literals


from pygments.lexer import RegexLexer, bygroups, words
from pygments.token import Comment, Operator, Keyword, Name, String, Number, Punctuation, Whitespace

__all__ = ['CrmshLexer']


class CrmshLexer(RegexLexer):
    """
    Lexer for `crmsh <http://crmsh.github.io/>`_ configuration files
    for Pacemaker clusters.

    .. versionadded:: 2.1
    """
    name = 'Crmsh'
    aliases = ['crmsh', 'pcmk']
    filenames = ['*.crmsh', '*.pcmk']
    mimetypes = []

    elem = words((
        'node', 'primitive', 'group', 'clone', 'ms', 'location',
        'colocation', 'order', 'fencing_topology', 'rsc_ticket',
        'rsc_template', 'property', 'rsc_defaults',
        'op_defaults', 'acl_target', 'acl_group', 'user', 'role',
        'tag'), suffix=r'(?![\w#$-])')
    sub = words((
        'params', 'meta', 'operations', 'op', 'rule',
        'attributes', 'utilization'), suffix=r'(?![\w#$-])')
    acl = words(('read', 'write', 'deny'), suffix=r'(?![\w#$-])')
    bin_rel = words(('and', 'or'), suffix=r'(?![\w#$-])')
    un_ops = words(('defined', 'not_defined'), suffix=r'(?![\w#$-])')
    date_exp = words(('in_range', 'date', 'spec', 'in'), suffix=r'(?![\w#$-])')
    acl_mod = (r'(?:tag|ref|reference|attribute|type|xpath)')
    bin_ops = (r'(?:lt|gt|lte|gte|eq|ne)')
    val_qual = (r'(?:string|version|number)')
    rsc_role_action = (r'(?:Master|Started|Slave|Stopped|'
                       r'start|promote|demote|stop)')

    tokens = {
        'root': [
            (r'^#.*\n?', Comment),
            # attr=value (nvpair)
            (r'([\w#$-]+)(=)("(?:""|[^"])*"|\S+)',
                bygroups(Name.Attribute, Punctuation, String)),
            # need this construct, otherwise numeric node ids
            # are matched as scores
            # elem id:
            (r'(node)(\s+)([\w#$-]+)(:)',
                bygroups(Keyword, Whitespace, Name, Punctuation)),
            # scores
            (r'([+-]?([0-9]+|inf)):', Number),
            # keywords (elements and other)
            (elem, Keyword),
            (sub, Keyword),
            (acl, Keyword),
            # binary operators
            (r'(?:%s:)?(%s)(?![\w#$-])' % (val_qual, bin_ops),
                Operator.Word),
            # other operators
            (bin_rel, Operator.Word),
            (un_ops, Operator.Word),
            (date_exp, Operator.Word),
            # builtin attributes (e.g. #uname)
            (r'#[a-z]+(?![\w#$-])', Name.Builtin),
            # acl_mod:blah
            (r'(%s)(:)("(?:""|[^"])*"|\S+)' % acl_mod,
                bygroups(Keyword, Punctuation, Name)),
            # rsc_id[:(role|action)]
            # NB: this matches all other identifiers
            (r'([\w#$-]+)(?:(:)(%s))?(?![\w#$-])' % rsc_role_action,
                bygroups(Name, Punctuation, Operator.Word)),
            # punctuation
            (r'(\\(?=\n)|[[\](){}/:@])', Punctuation),
            (r'\s+|\n', Whitespace),
        ],
    }



# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

# Helper completers

from . import xmlutil


def choice(lst):
    '''
    Static completion from a list
    '''
    def completer(args):
        return lst
    return completer


null = choice([])
attr_id = choice(["id="])

def call(fn, *fnargs):
    '''
    Call the given function with the given arguments.
    The function has to return a list of completions.
    '''
    def completer(args):
        return fn(*fnargs)
    return completer


def join(*fns):
    '''
    Combine the output of several completers
    into a single completer.
    '''
    def completer(args):
        ret = []
        for fn in fns:
            ret += list(fn(args))
        return ret
    return completer


booleans = choice(['yes', 'no', 'true', 'false', 'on', 'off'])


def resources(args=None):
    cib_el = xmlutil.resources_xml()
    if cib_el is None:
        return []
    nodes = xmlutil.get_interesting_nodes(cib_el, [])
    rsc_id_list = [x.get("id") for x in nodes if xmlutil.is_resource(x)]
    if args:
        if args[0] in ('promote', 'demote'):
            rsc_id_list = [item for item in rsc_id_list if xmlutil.RscState().is_ms_or_promotable_clone(item)]
        elif args[0] == "start":
            rsc_id_list = [item for item in rsc_id_list if not xmlutil.RscState().is_running(item)]
        elif args[0] == "stop":
            rsc_id_list = [item for item in rsc_id_list if xmlutil.RscState().is_running(item)]
        rsc_id_list = [item for item in rsc_id_list if item not in args]
    return rsc_id_list


def primitives(args):
    cib_el = xmlutil.resources_xml()
    if cib_el is None:
        return []
    nodes = xmlutil.get_interesting_nodes(cib_el, [])
    return [x.get("id") for x in nodes if xmlutil.is_primitive(x)]


nodes = call(xmlutil.listnodes)
online_nodes = call(lambda x: xmlutil.CrmMonXmlParser().get_node_list(standby=x), False)
standby_nodes = call(lambda x: xmlutil.CrmMonXmlParser().get_node_list(standby=x), True)

shadows = call(xmlutil.listshadows)



contents = {}

contents["sbd"]= '''On {nodeA}, once the sbd process get killed, there are two situations:
  a) sbd process restarted
     Systemd will restart sbd service immediately.
     Restarting sbd service will also lead to restart corosync and pacemaker services because of the pre-defined dependencies among the systemd unit files.

  b) {nodeA} experience the watchdog fencing
     There is the race condition with the watchdog timer. Watchdog might reset {nodeA}, just before the sbd service get restarted and not tickle the watchdog timer in time.'''

contents["sbd-l"] = '''On {nodeA}, the sbd service is killed consistantly all the time. 
Very quickly, systemd will hit the start limit to restart sbd service. 
Basically, in the end, systemd stops restarting anymore, marks the sbd service as failure.
{nodeB} sbd cluster health check marks it as "UNHEALTHY".
{nodeB} treats {nodeA} as a node lost, and fences it in the end.'''

contents["corosync"] = '''On {nodeA}, once the corosync process get killed, systemd will restart corosync service immediately. There are two situations:
  a) corosync process restarts
     {nodeA} corosync process get restarted and rejoins to the existent membership quickly enough.
     Basically, it happens before {nodeB} treats it as a node lost.
     In the end, the cluster looks like nothing happened to the user. RA stays safe and sound.

  b) {nodeA} gets fenced
     {nodeA} gets fenced since {nodeB} corosync just ran out of timeout and treat it as a node lost and forms a new membership.
     The decision making process of {nodeB}, pengine(aka. schedulerd in Pacemaker 2), will initiate fence action against {nodeA}. '''

contents["corosync-l"] = '''The corosync service is killed consistantly all the time.
Very quickly, systemd will hit the start limit to restart corosync service.
Basically, in the end, systemd stops restarting anymore, marks the corosync service as failure. {nodeB} treats {nodeA} as a node lost, marks it as "unclean", and fence it in the end.'''

contents["pacemakerd"] = '''The pacemakerd process gets restarted by systemd. All RAs must stay intact.'''



import re
import unittest
from unittest import mock

import lxml.etree

from crmsh import migration, cibquery


class TestCheckRemovedResourceAgents(unittest.TestCase):
    def setUp(self):
        self._handler = mock.Mock(migration.CheckResultHandler)

    def test_load_unsupported_resource_agents(self):
        s = migration.UnsupportedResourceAgentDetector()
        self.assertEqual(
            migration.UnsupportedResourceAgentDetector.UnsupportedState(
                cibquery.ResourceAgent('ocf', 'heartbeat', 'IPaddr2'),
                False,
            ),
            s.get_unsupported_state(cibquery.ResourceAgent('ocf', 'heartbeat', 'IPaddr'))
        )
        self.assertEqual(
            migration.UnsupportedResourceAgentDetector.UnsupportedState(
                cibquery.ResourceAgent('stonith', None, 'fence_sbd'),
                False,
            ),
            s.get_unsupported_state(cibquery.ResourceAgent('stonith', None, 'external/sbd'))
        )
        self.assertEqual(
            migration.UnsupportedResourceAgentDetector.UnsupportedState(
                None,
                False,
            ),
            s.get_unsupported_state(cibquery.ResourceAgent('ocf', 'heartbeat', 'rkt'))
        )
        self.assertEqual(
            migration.UnsupportedResourceAgentDetector.UnsupportedState(
                cibquery.ResourceAgent('ocf', 'heartbeat', 'LVM-activate'),
                True,
            ),
            s.get_unsupported_state(cibquery.ResourceAgent('ocf', 'heartbeat', 'LVM'))
        )

    def test_check_removed_resource_agents(self):
        mock_detector = mock.Mock(migration.UnsupportedResourceAgentDetector)
        mock_detector.get_unsupported_state.side_effect = [
            migration.UnsupportedResourceAgentDetector.UnsupportedState(cibquery.ResourceAgent('foo', 'bar', 'qux2'), True),
            migration.UnsupportedResourceAgentDetector.UnsupportedState(None, False),
        ]
        migration._check_removed_resource_agents(
            self._handler,
            'msg',
            mock_detector,
            [
                cibquery.ResourceAgent('foo', 'bar', 'qux'),
                cibquery.ResourceAgent('a', 'b', 'c'),
            ]
        )
        self._handler.handle_problem.assert_called()

    def test_check_version_range(self):
        def check_fn(x):
            migration._check_version_range(
                self._handler,
                'foo',
                (1, 1,),
                re.compile(r'^foo\s+(\d+(?:.\d+)*)'),
                x,
            )
        check_fn('foo 0')
        self._handler.handle_problem.assert_called()
        self._handler.handle_problem.reset_mock()
        check_fn('foo 0.9')
        self._handler.handle_problem.assert_called()
        self._handler.handle_problem.reset_mock()
        check_fn('foo 0.9.99')
        self._handler.handle_problem.assert_called()
        self._handler.handle_problem.reset_mock()
        check_fn('foo 1')
        self._handler.handle_problem.assert_called()
        self._handler.handle_problem.reset_mock()
        check_fn('foo 1.1')
        self._handler.handle_problem.assert_not_called()
        check_fn('foo 1.1.0')
        self._handler.handle_problem.assert_not_called()
        check_fn('foo 1.1.1')
        self._handler.handle_problem.assert_not_called()
        check_fn('foo 1.2')
        self._handler.handle_problem.assert_not_called()
        check_fn('foo 2')
        self._handler.handle_problem.assert_not_called()
        check_fn('foo 2.0')
        self._handler.handle_problem.assert_not_called()

    @mock.patch('glob.iglob')
    def test_get_latest_cib_schema_version(self, mock_iglob: mock.MagicMock):
        mock_iglob.return_value = iter([
            'pacemaker-0.1.rng', 'pacemaker-1.9.rng', 'pacemaker-1.11.rng', 'pacemaker-next.rng',
        ])
        self.assertEqual((1, 11), migration._get_latest_cib_schema_version())

    @mock.patch('crmsh.migration._get_latest_cib_schema_version')
    def test_check_cib_schema_version(self, mock_get_latest_cib_schema_version):
        cib = lxml.etree.fromstring('<cib crm_feature_set="3.16.1" validate-with="pacemaker-3.9" epoch="7" num_updates="0" admin_epoch="0" cib-last-written="Fri Jan  3 13:35:49 2025" update-origin="ha-1-2" update-client="cibadmin" update-user="root" have-quorum="1" dc-uuid="1"/>')
        mock_get_latest_cib_schema_version.return_value = (3, 10)
        handler = mock.Mock(migration.CheckResultHandler)
        migration.check_cib_schema_version(handler, cib)
        handler.handle_problem.assert_called_with(
            False, False, handler.LEVEL_WARN,
            "The CIB is not validated with the latest schema version.", [
                '* Latest version:  3.10',
                '* Current version: 3.9',
                'Please run "crm configure upgrade force" to upgrade to the latest version.',
            ]
        )



# Tool to sort the documentation alphabetically
# Makes a lot of assumptions about the structure of the document it edits
# Looks for special markers that indicate structure

# prints output to stdout

# print lines until in a cmdhelp_<section>
# collect all cmdhelp_<section>_<subsection> subsections
# sort and print

import sys
import re


class Sorter(object):
    def __init__(self):
        self.current_section = None
        self.current_subsection = None
        self.subsections = []
        self.re_section = re.compile(r'^\[\[cmdhelp_([^_,]+),')
        self.re_subsection = re.compile(r'^\[\[cmdhelp_([^_]+)_([^,]+),')

    def beginsection(self, line):
        m = self.re_section.match(line)
        name = m.group(1)
        self.current_section = [name, line]
        self.current_subsection = None
        self.subsections = []
        return self.insection

    def insection(self, line):
        if line.startswith('[[cmdhelp_%s_' % (self.current_section[0])):
            return self.beginsubsection(line)
        elif line.startswith('[['):
            self.finishsection()
            return self.preprint(line)
        else:
            self.current_section[1] += line
        return self.insection

    def beginsubsection(self, line):
        m = self.re_subsection.match(line)
        name = m.group(2)
        self.current_subsection = [name, line]
        return self.insubsection

    def insubsection(self, line):
        if line.startswith('[['):
            self.subsections.append(self.current_subsection)
            self.current_subsection = None
            return self.insection(line)
        self.current_subsection[1] += line
        return self.insubsection

    def finishsection(self):
        if self.current_section:
            print self.current_section[1],
            for name, text in sorted(self.subsections, key=lambda x: x[0]):
                print text,
        self.current_section = None
        self.subsections = []

    def preprint(self, line):
        if self.re_section.match(line):
            return self.beginsection(line)
        print line,
        return self.preprint

    def run(self, lines):
        action = self.preprint
        for line in lines:
            prevaction = action
            action = action(line)
            if action is None:
                print prevaction
                print self.current_section
                print self.current_subsection
                sys.exit(1)
        if self.current_section:
            self.finishsection()

Sorter().run(open(sys.argv[1]).readlines())



import os
import time
import threading
import shutil
import tempfile
from datetime import datetime
from contextlib import contextmanager
from crmsh import utils as crmshutils
from crmsh import xmlutil
from crmsh import log
from . import utils
from . import config
from ..service_manager import ServiceManager
from ..sh import ShellUtils

logger = log.setup_logger(__name__)


class TaskError(Exception):
    pass


class Task(object):
    """
    Task is a base class
    Use for record the information of each test case
    """
    REBOOT_WARNING = """!!! WARNING WARNING WARNING !!!
THIS CASE MAY LEAD TO NODE BE FENCED.
TYPE Yes TO CONTINUE, OTHER INPUTS WILL CANCEL THIS CASE [Yes/No](No): """
    TIME_STR_FORMAT = '%Y/%m/%d %H:%M:%S'

    def __init__(self, description, flush=False, quiet=False):
        """
        Init function
        flush, to print the message immediately
        """
        self.passed = True
        self.force = False
        self.quiet = quiet
        self.messages = []
        self.timestamp = datetime.now()
        self.description = description
        utils.msg_info(self.description, to_stdout=False)
        self.flush = flush
        self.fence_start_event = threading.Event()
        self.fence_finish_event = threading.Event()
        self.thread_stop_event = threading.Event()
        from . import main
        self.prev_task_list = main.ctx.task_list

    def info(self, msg):
        self.msg_append("info", msg)
        utils.msg_info(msg, to_stdout=self.flush)

    def warn(self, msg):
        self.msg_append("warn", msg)
        utils.msg_warn(msg, to_stdout=self.flush)

    def error(self, msg):
        self.msg_append("error", msg)
        utils.msg_error(msg, to_stdout=self.flush)

    def msg_append(self, msg_type, msg):
        if msg_type == "error":
            self.passed = False
        self.messages.append((msg_type, msg, utils.now()))
        if self.flush:
            self.to_json()
            self.to_report()

    def header(self):
        pass

    def to_report(self):
        pass

    def to_json(self):
        pass

    def build_base_result(self):
        """
        Build base results
        """
        self.result = {
            "Timestamp": self.timestamp.strftime(self.TIME_STR_FORMAT),
            "Description": self.description,
            "Messages": ["{} {}:{}".format(m[2], m[0].upper(), m[1])
                         for m in self.messages]
        }

    def print_header(self):
        """
        Print testcase header
        """
        print(self.header())
        if not self.force and not utils.warning_ask(self.REBOOT_WARNING):
            self.info("Testcase cancelled")
            raise crmshutils.TerminateSubCommand

    def task_pre_check(self, need_fence=True):
        """
        Prerequisite check
          * pacemaker.service is active
          * stonith is enabled
        """
        if not ServiceManager().service_is_active("pacemaker.service"):
            raise TaskError("Cluster not running!")
        if need_fence:
            self.get_fence_info()
            if not self.fence_enabled:
                raise TaskError("Require stonith enabled")
            if not self.fence_configured:
                raise TaskError("Require stonith device configured and running")


    def get_fence_info(self):
        """
        Get fence info
        """
        fence_info_inst = utils.FenceInfo()
        self.fence_enabled = fence_info_inst.fence_enabled
        self.fence_configured = fence_info_inst.fence_configured
        self.fence_action = fence_info_inst.fence_action
        self.fence_timeout = fence_info_inst.fence_timeout

    def fence_action_monitor(self):
        """
        Monitor fencing process, running in thread, exit on two cases:
        1. There is one latest fence action successfully done
        2. No fence action during fence timeout, thread_stop_event triggered by main thread
        """
        # Try to find out which node fire the fence action
        while not self.thread_stop_event.is_set():
            fence_event_dict = xmlutil.CrmMonXmlParser().get_last_fence_event_info()
            if fence_event_dict:
                target_node = fence_event_dict.get('target')
                origin_node = fence_event_dict.get('origin')
                complete_time = fence_event_dict.get('completed')
                status = fence_event_dict.get('status')
                if status == "pending" and not self.fence_start_event.is_set():
                    self.info(f"Node \"{target_node}\" will be fenced by \"{origin_node}\"!")
                    self.fence_start_event.set()
                # Try to find out proof that fence happened
                elif status == "success":
                    task_timestamp = self.timestamp.timestamp()
                    complete_timestamp = datetime.fromisoformat(complete_time).timestamp()
                    # This success event should after the task started
                    if task_timestamp < complete_timestamp:
                        self.info(f"Node \"{target_node}\" was fenced by \"{origin_node}\" at {complete_time}")
                        self.fence_finish_event.set()
                        break
            time.sleep(1)


class TaskFence(Task):
    """
    Class to fence node
    """
    def  __init__(self, context):
        """
        Init function
        """
        self.target_node = context.fence_node
        description = "Fence node {}".format(self.target_node)
        super(self.__class__, self).__init__(description, flush=True)
        self.force = context.force

    def header(self):
        """
        Header content for this task
        """
        h = '''==============================================
Testcase:          {}
Fence action:      {}
Fence timeout:     {}
'''.format(self.description, self.fence_action, self.fence_timeout)
        return h

    def to_json(self):
        """
        Dump join result
        """
        self.build_base_result()
        self.result['Fence action'] = self.fence_action
        self.result['Fence timeout'] = self.fence_timeout
        from . import main
        main.ctx.task_list = self.prev_task_list + [self.result]
        utils.json_dumps()

    def pre_check(self):
        """
        Check the prerequisite for fence node
        """
        self.task_pre_check()

        for cmd in ['crm_node', 'stonith_admin', 'crm_attribute']:
            rc, _, err = ShellUtils().get_stdout_stderr("which {}".format(cmd))
            if rc != 0 and err:
                raise TaskError(err)

        if not utils.check_node_status(self.target_node, 'member'):
            raise TaskError("Node \"{}\" not in cluster!".format(self.target_node))

    def run(self):
        """
        Fence node and start a thread to monitor the result
        """
        self.info("Trying to fence node \"{}\"".format(self.target_node))
        ShellUtils().get_stdout_stderr(config.FENCE_NODE.format(self.target_node))
        th = threading.Thread(target=self.fence_action_monitor)
        th.start()

    def wait(self):
        """
        Wait until fence happened
        """
        if self.target_node == utils.this_node():
            self.info("Waiting {}s for self {}...".format(self.fence_timeout, self.fence_action))
        else:
            self.info("Waiting {}s for node \"{}\" {}...".format(self.fence_timeout,
                self.target_node, self.fence_action))

        result = self.fence_finish_event.wait(int(self.fence_timeout))
        if not result:
            self.thread_stop_event.set()
            raise TaskError("Target fence node \"{}\" still alive".format(self.target_node))


class TaskCheck(Task):
    """
    Class to define the format of output for checking item results and how to dump json
    """

    def __init__(self, description, quiet=False):
        """
        Init function
        """
        super(self.__class__, self).__init__(description, quiet=quiet)

    def to_stdout(self):
        """
        Define the format of results to stdout
        """
        with utils.manage_handler("file", keep=False):
            utils.get_handler(logger, "stream").setFormatter(utils.MyLoggingFormatter(flush=False))

            if self.passed:
                message = "{} [{}]".format(self.description, utils.CGREEN + "Pass" + utils.CEND)
            else:
                message = "{} [{}]".format(self.description, utils.CRED + "Fail" + utils.CEND)
            logger.info(message, extra={'timestamp': '[{}]'.format(self.timestamp.strftime(self.TIME_STR_FORMAT))})

            for msg in self.messages:
                logger.log(utils.LEVEL[msg[0]], msg[1], extra={'timestamp': '  '})

            utils.get_handler(logger, "stream").setFormatter(utils.MyLoggingFormatter())

    def to_json(self):
        """
        Json results
        """
        self.build_base_result()
        self.result['Result'] = self.passed
        from . import main
        main.ctx.task_list.append(self.result)
        utils.json_dumps()

    def print_result(self):
        """
        Print results to stdout and json
        """
        if self.quiet:
            return
        self.to_stdout()
        self.to_json()

    @contextmanager
    def run(self):
        """
        Context manager to do things and print results finally
        """
        try:
            yield
        finally:
            self.print_result()


class TaskKill(Task):
    """
    Class to define how to run kill testcases
    """

    EXPECTED = {
        # process_name: (expected_results, expected_results_with_loop)
        'sbd':        ('''a) sbd process restarted
                   b) Or, this node fenced.''', 'This node fenced'),
        'corosync':   ('''a) corosync process restarted
                   b) Or, this node fenced.''', 'This node fenced'),
        'pacemakerd': ('pacemakerd process restarted', None),
    }
    WAIT_TIMEOUT = 10

    def  __init__(self, context):
        """
        Init function
        """
        self.target_kill = context.current_case
        self.description = "Force kill {}".format(self.target_kill)
        super(self.__class__, self).__init__(self.description, flush=True)
        self.cmd = "killall -9 {}".format(self.target_kill)
        self.looping = context.loop
        self.force = context.force
        if not self.looping:
            self.expected = self.EXPECTED[self.target_kill][0]
        else:
            self.expected = self.EXPECTED[self.target_kill][1]
        self.report = False
        self.restart_happen_event = threading.Event()

    def enable_report(self):
        """
        Enable report
        """
        self.report = True
        from . import main
        if not os.path.isdir(main.ctx.report_path):
            raise TaskError("{} is not a directory".format(main.ctx.report_path))

        report_path = main.ctx.report_path
        report_name = "{}-{}.report".format(main.ctx.process_name, utils.now("%Y%m%d-%s"))
        self.report_file = os.path.join(report_path, report_name)
        print("(Report: {})".format(self.report_file))

        if self.looping:
            content_key = "{}-l".format(self.target_kill)
        else:
            content_key = self.target_kill

        from . import explain
        self.explain = explain.contents[content_key].format(nodeA=utils.this_node(), nodeB="other node")

    def header(self):
        """
        Define descriptions
        """
        h = '''==============================================
Testcase:          {}
Looping Kill:      {}
Expected State:    {}
'''.format(self.description, self.looping, self.expected)
        return h

    def to_json(self):
        """
        Json results
        """
        self.build_base_result()
        self.result['Looping Kill'] = self.looping
        self.result['Expected State'] = self.expected
        from . import main
        main.ctx.task_list = self.prev_task_list + [self.result]
        utils.json_dumps()

    def to_report(self):
        """
        Generate report
        """
        if not self.report:
            return
        with open(self.report_file, 'w') as f:
            f.write(self.header())
            f.write("\nLog:\n")
            for m in self.messages:
                f.write("{} {}:{}\n".format(m[2], m[0].upper(), m[1]))
            f.write("\nTestcase Explained:\n")
            f.write("{}\n".format(self.explain))
            f.flush()
            os.fsync(f)

    def pre_check(self):
        """
        Check the prerequisite
        """
        self.task_pre_check()
        rc, pid = utils.get_process_status(self.target_kill)
        if not rc:
            raise TaskError("Process {} is not running!".format(self.target_kill))

    def run(self):
        """
        Execute specific kill command and monitor the results
        """
        while True:
            rc, pid = utils.get_process_status(self.target_kill)
            if rc:
                self.info("Process {}({}) is running...".format(self.target_kill, pid))
            else:
                continue
            self.info("Trying to run \"{}\"".format(self.cmd))
            ShellUtils().get_stdout_stderr(self.cmd)
            # endless loop will lead to fence
            if not self.looping:
                break

        fence_check_th = threading.Thread(target=self.fence_action_monitor)
        fence_check_th.start()
        restart_check_th = threading.Thread(target=self.process_monitor)
        restart_check_th.start()

    def wait(self):
        """
        Wait process to restart
        """
        if self.fence_start_event.wait(self.WAIT_TIMEOUT) and not self.restart_happen_event.is_set():
            raise TaskError("Process {} is not restarted!".format(self.target_kill))
        self.thread_stop_event.set()

    def process_monitor(self):
        """
        Monitor process status
        """
        while not self.thread_stop_event.is_set():
            rc, pid = utils.get_process_status(self.target_kill)
            if rc:
                self.info("Process {}({}) is restarted!".format(self.target_kill, pid))
                self.restart_happen_event.set()
                break
            time.sleep(1)


class TaskSplitBrain(Task):
    """
    Class to define how to simulate split brain by blocking traffic between cluster nodes
    """

    def  __init__(self, force=False):
        """
        Init function
        """
        self.description = "Simulate split brain by blocking traffic between cluster nodes"
        self.expected = "One of nodes get fenced"
        self.ports = []
        self.peer_nodelist = []
        super(self.__class__, self).__init__(self.description, flush=True)
        self.force = force

    def header(self):
        """
        Define descriptions
        """
        h = '''==============================================
Testcase:          {}
Expected Result:   {}
Fence action:      {}
Fence timeout:     {}
'''.format(self.description, self.expected, self.fence_action, self.fence_timeout)
        return h

    def to_json(self):
        """
        Json results
        """
        self.build_base_result()
        self.result['Fence action'] = self.fence_action
        self.result['Fence timeout'] = self.fence_timeout
        from . import main
        main.ctx.task_list = self.prev_task_list + [self.result]
        utils.json_dumps()

    def pre_check(self):
        """
        Check the prerequisite
        """
        self.task_pre_check()

        for cmd in ["iptables"]:
            rc, _, err = ShellUtils().get_stdout_stderr("which {}".format(cmd))
            if rc != 0 and err:
                raise TaskError(err)

        if len(utils.online_nodes()) < 2:
            raise TaskError("At least two nodes online!")

    @contextmanager
    def do_block(self):
        """
        Context manager to block and unblock ip/ports
        """
        self.do_block_iptables()
        try:
            yield
        finally:
            self.un_block()

    def do_block_iptables(self):
        """
        Block corosync communication ip
        """
        self.peer_nodelist = utils.peer_node_list()
        for node in self.peer_nodelist:
            self.info("Trying to temporarily block {} communication ip".format(node))
            for ip in crmshutils.get_iplist_from_name(node):
                ShellUtils().get_stdout_stderr(config.BLOCK_IP.format(action='I', peer_ip=ip))

    def un_block(self):
        """
        Unblock corosync ip/ports
        """
        self.un_block_iptables()

    def un_block_iptables(self):
        """
        Unblock corosync communication ip
        """
        for node in self.peer_nodelist:
            self.info("Trying to recover {} communication ip".format(node))
            for ip in crmshutils.get_iplist_from_name(node):
                ShellUtils().get_stdout_stderr(config.BLOCK_IP.format(action='D', peer_ip=ip))

    def run(self):
        """
        Fence node and start a thread to monitor the result
        """
        #self.info("Trying to fence node \"{}\"".format(self.target_node))
        #ShellUtils().get_stdout_stderr(config.FENCE_NODE.format(self.target_node), wait=False)
        th = threading.Thread(target=self.fence_action_monitor)
        th.start()

    def wait(self):
        """
        Wait until fence happened
        """
        result = self.fence_finish_event.wait(int(self.fence_timeout))
        if not result:
            self.thread_stop_event.set()
            # should be an error here


class TaskFixSBD(Task):
    """
    Class to fix SBD DEVICE incorrect issue
    """

    def  __init__(self, candidate, force=False):
        self.new = candidate
        self.description = "Replace SBD_DEVICE with candidate {}".format(self.new)
        self.conf = config.SBD_CONF
        super(self.__class__, self).__init__(self.description, flush=True)
        self.bak = tempfile.mkstemp()[1]
        self.edit = tempfile.mkstemp()[1]
        self.force = force

        sbd_options = crmshutils.parse_sysconfig(self.conf)
        self.old = sbd_options["SBD_DEVICE"]

    def header(self):
        """
        Case header
        """
        h = '''==============================================
Case:                {}
Original SBD device: {}
New SBD device:      {}
'''.format(self.description, self.old, self.new)
        return h

    def to_json(self):
        """
        Generate json output
        """
        self.build_base_result()
        self.result['Original SBD device'] = self.old
        self.result['New SBD device'] = self.new
        from . import main
        main.ctx.task_list = self.prev_task_list + [self.result]
        utils.json_dumps()

    def pre_check(self):
        """
        Check the prerequisite
        """
        if not os.path.exists(self.conf):
            raise TaskError("Configure file {} not exist!".format(self.conf))

        if not os.path.exists(self.new):
            raise TaskError("Device {} not exist!".format(self.new))

    @contextmanager
    def backup(self):
        """
        Backup the configuration file before modify
        """
        shutil.copyfile(self.conf, self.bak)
        try:
            yield
        finally:
            if self.bak:
                shutil.copyfile(self.bak, self.conf)

    def run(self):
        """
        Change the SBD DEVICE of configuration file
        """
        with open(self.edit, "w") as editfd:
            with open(self.conf, "r") as oldfd:
                for line in oldfd.readlines():
                    if line.strip().startswith("SBD_DEVICE"):
                        line = "SBD_DEVICE='" + self.new +"'\n"
                    editfd.write(line)

        try:
            shutil.copymode(self.conf, self.edit)
            os.remove(self.conf)
            shutil.move(self.edit, self.conf)
            os.remove(self.bak)
            self.bak = None
        except:
            raise TaskError("Fail to modify file {}".format(self.conf))

    def verify(self):
        """
        Verify the modification is working
        """
        sbd_options = crmshutils.parse_sysconfig(self.conf)

        if sbd_options["SBD_DEVICE"] == self.new:
            self.info("SBD DEVICE change succeed")
        else:
            raise TaskError("Fail to replace SBD device {} in {}!".
                            format(self.new, config.SBD_CONF))



import subprocess
from crmsh import config
import crmsh.report.sh
from crmsh.report import core, constants, utils, collect
import crmsh.log

import sys
import argparse
import unittest
from unittest import mock


class TestCapitalizedHelpFormatter(unittest.TestCase):
    def setUp(self):
        # Initialize the ArgumentParser with the CapitalizedHelpFormatter
        self.parser = argparse.ArgumentParser(
            formatter_class=core.CapitalizedHelpFormatter,
            usage="usage: test"
        )
        self.parser.add_argument('--test', help='Test option')

    def test_usage(self):
        # Test that the usage is capitalized
        usage_text = self.parser.format_usage()
        self.assertTrue(usage_text.startswith('Usage: '))

    def test_section_heading(self):
        # Test that section headings are capitalized
        section_text = self.parser.format_help()
        self.assertTrue('Option' in section_text)


class TestContext(unittest.TestCase):

    @mock.patch('crmsh.report.utils.parse_to_timestamp')
    @mock.patch('crmsh.report.utils.now')
    @mock.patch('crmsh.report.core.config')
    def setUp(self, mock_config, mock_now, mock_parse_to_timestamp):
        mock_config.report = mock.Mock(
            from_time="20230101",
            compress=False,
            collect_extra_logs=["file1", "file2"],
            remove_exist_dest=False,
            single_node=False
        )
        mock_now.return_value = "12345"
        mock_parse_to_timestamp.return_value = "54321"
        self.context = core.Context()
        self.context.load()

    def test_attribute_setting(self):
        self.context.name = "value"
        self.assertEqual(self.context.name, "value")
        self.context["age"] = 19
        self.assertEqual(self.context.age, 19)
        self.context.extra_log_list = ["file3", "file2"]
        self.assertEqual(len(self.context.extra_log_list), 3)

    @mock.patch('json.dumps')
    def test_str(self, mock_dumps):
        mock_dumps.return_value = "json str"
        self.assertEqual(self.context.name, "crm_report")
        self.assertEqual(self.context.from_time, "54321")
        self.assertEqual(str(self.context), "json str")


class TestRun(unittest.TestCase):

    @mock.patch('os.path.isdir')
    def test_process_dest_dest_not_exist(self, mock_isdir):
        mock_isdir.return_value = False
        mock_ctx_inst = mock.Mock(dest="/opt/test/report")
        with self.assertRaises(utils.ReportGenericError) as err:
            core.process_dest(mock_ctx_inst)
        self.assertEqual("Directory /opt/test does not exist", str(err.exception))

    @mock.patch('crmsh.utils.is_filename_sane')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.isdir')
    def test_process_dest_filename_not_sane(self, mock_isdir, mock_basename, mock_sane):
        mock_isdir.return_value = True
        mock_sane.return_value = False
        mock_basename.return_value = "report*"
        mock_ctx_inst = mock.Mock(dest="/opt/test/report*")
        with self.assertRaises(utils.ReportGenericError) as err:
            core.process_dest(mock_ctx_inst)
        self.assertEqual("report* is invalid file name", str(err.exception))

    @mock.patch('crmsh.report.core.pick_compress_prog')
    @mock.patch('shutil.rmtree')
    @mock.patch('crmsh.utils.is_filename_sane')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.isdir')
    def test_process_dest_dir_exists_rmtree(self, mock_isdir, mock_basename, mock_sane, mock_rmtree, mock_pick):
        mock_isdir.side_effect = [True, True]
        mock_sane.return_value = True
        mock_basename.return_value = "report"
        mock_ctx_inst = mock.Mock(dest="/opt/test/report", no_compress=True, rm_exist_dest=True)
        core.process_dest(mock_ctx_inst)
        mock_rmtree.assert_called_once_with("/opt/test/report")

    @mock.patch('crmsh.report.core.pick_compress_prog')
    @mock.patch('crmsh.utils.is_filename_sane')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.isdir')
    def test_process_dest_dir_exists(self, mock_isdir, mock_basename, mock_sane, mock_pick):
        mock_isdir.side_effect = [True, True]
        mock_sane.return_value = True
        mock_basename.return_value = "report"
        mock_ctx_inst = mock.Mock(dest="/opt/test/report", no_compress=True, rm_exist_dest=False)
        with self.assertRaises(utils.ReportGenericError) as err:
            core.process_dest(mock_ctx_inst)
        self.assertEqual("Destination directory /opt/test/report exists, please cleanup or use -Z option", str(err.exception))

    @mock.patch('crmsh.report.core.pick_compress_prog')
    @mock.patch('crmsh.utils.is_filename_sane')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.isdir')
    @mock.patch('crmsh.report.utils.now')
    def test_process_dest(self, mock_now, mock_isdir, mock_basename, mock_is_sane, mock_pick):
        mock_now.return_value = "Mon-28-Aug-2023"
        mock_isdir.side_effect = [True, False]
        mock_is_sane.return_value = True
        mock_basename.return_value = f"report.{mock_now.return_value}"
        mock_ctx_inst = mock.Mock(dest=None, no_compress=False, compress_suffix=".bz2", name="report")

        core.process_dest(mock_ctx_inst)

        self.assertEqual(mock_ctx_inst.dest_dir, ".")
        mock_is_sane.assert_called_once_with(mock_basename.return_value)
        self.assertEqual(mock_ctx_inst.dest_path, "./report.Mon-28-Aug-2023.tar.bz2")

    @mock.patch('crmsh.report.core.pick_first_compress')
    def test_pick_compress_prog(self, mock_pick):
        mock_pick.return_value = (None, None)
        mock_ctx_inst = mock.Mock()
        core.pick_compress_prog(mock_ctx_inst)
        self.assertEqual(mock_ctx_inst.compress_prog, "cat")

    @mock.patch('shutil.which')
    def test_pick_first_compress_return(self, mock_which):
        mock_which.return_value = True
        prog, ext = core.pick_first_compress()
        self.assertEqual(prog, "gzip")
        self.assertEqual(ext, ".gz")
        mock_which.assert_called_once_with("gzip")

    @mock.patch('logging.Logger.warning')
    @mock.patch('shutil.which')
    def test_pick_first_compress(self, mock_which, mock_warn):
        mock_which.side_effect = [False, False, False, False]
        prog, ext = core.pick_first_compress()
        self.assertIsNone(prog)
        self.assertIsNone(ext)

    @mock.patch('crmsh.report.utils.get_timespan_str')
    @mock.patch('logging.Logger.info')
    def test_finalword(self, mock_info, mock_get_timespan):
        mock_ctx_inst = mock.Mock(dest_path="./crm_report-Tue-15-Aug-2023.tar.bz2", node_list=["node1", "node2"])
        mock_get_timespan.return_value = "2023-08-14 18:17 - 2023-08-15 06:17"
        core.finalword(mock_ctx_inst)
        mock_info.assert_has_calls([
            mock.call(f"The report is saved in {mock_ctx_inst.dest_path}"),
            mock.call(f"Report timespan: {mock_get_timespan.return_value}"),
            mock.call(f"Including nodes: {' '.join(mock_ctx_inst.node_list)}"),
            mock.call("Thank you for taking time to create this report")
            ])

    @mock.patch('os.path.basename')
    @mock.patch('crmsh.report.core.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.mkdirp')
    @mock.patch('crmsh.report.core.is_collector')
    @mock.patch('crmsh.report.core.tmpfiles.create_dir')
    def test_setup_workdir_collector(self, mock_create_dir, mock_collector, mock_mkdirp, mock_logger, mock_basename):
        mock_create_dir.return_value = "/tmp/tmp_dir"
        mock_ctx_inst = mock.Mock(dest="/opt/report", work_dir="/opt/work_dir", me="node1")
        mock_collector.return_value = True
        mock_basename.return_value = "report"
        core.setup_workdir(mock_ctx_inst)
        mock_logger.debug2.assert_called_once_with(f"Setup work directory in {mock_ctx_inst.work_dir}")

    @mock.patch('os.path.basename')
    @mock.patch('crmsh.report.core.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.mkdirp')
    @mock.patch('crmsh.report.core.is_collector')
    @mock.patch('crmsh.report.core.tmpfiles.create_dir')
    def test_setup_workdir(self, mock_create_dir, mock_collector, mock_mkdirp, mock_logger, mock_basename):
        mock_create_dir.return_value = "/tmp/tmp_dir"
        mock_ctx_inst = mock.Mock(dest="/opt/report", work_dir="/opt/work_dir")
        mock_collector.return_value = False
        mock_basename.return_value = "report"
        core.setup_workdir(mock_ctx_inst)
        mock_logger.debug2.assert_called_once_with(f"Setup work directory in {mock_ctx_inst.work_dir}")

    @mock.patch('os.path.isdir')
    @mock.patch('crmsh.report.core.load_from_crmsh_config')
    def test_load_context_attributes(self, mock_load, mock_isdir):
        mock_ctx_inst = mock.Mock(cib_dir="/var/lib/pacemaker/cib")
        mock_isdir.return_value = True

        core.load_context_attributes(mock_ctx_inst)

        self.assertEqual(mock_ctx_inst.pcmk_lib_dir, "/var/lib/pacemaker")
        self.assertEqual(mock_ctx_inst.cores_dir_list, ["/var/lib/pacemaker/cores", constants.COROSYNC_LIB])

    @mock.patch('os.path.isdir')
    @mock.patch('crmsh.report.core.config')
    def test_load_from_crmsh_config(self, mock_config, mock_isdir):
        mock_config.path = mock.Mock(
            crm_config="/var/lib/pacemaker/cib",
            crm_daemon_dir="/usr/lib/pacemaker",
            pe_state_dir="/var/lib/pacemaker/pe"
        )
        mock_isdir.side_effect = [True, True, True]
        mock_ctx_inst = mock.Mock()

        core.load_from_crmsh_config(mock_ctx_inst)

        self.assertEqual(mock_ctx_inst.cib_dir, mock_config.path.crm_config)
        self.assertEqual(mock_ctx_inst.pcmk_exec_dir, mock_config.path.crm_daemon_dir)
        self.assertEqual(mock_ctx_inst.pe_dir, mock_config.path.pe_state_dir)

    @mock.patch('os.path.isdir')
    @mock.patch('crmsh.report.core.config')
    def test_load_from_crmsh_config_exception(self, mock_config, mock_isdir):
        mock_config.path = mock.Mock(
            crm_config="/var/lib/pacemaker/cib",
        )
        mock_isdir.return_value = False
        mock_ctx_inst = mock.Mock()

        with self.assertRaises(utils.ReportGenericError) as err:
            core.load_from_crmsh_config(mock_ctx_inst)
        self.assertEqual(f"Cannot find CIB directory", str(err.exception))

    def test_adjust_verbosity_debug(self):
        mock_ctx_inst = mock.Mock(debug=1)
        core.adjust_verbosity(mock_ctx_inst)

    def test_adjust_verbosity(self):
        mock_ctx_inst = mock.Mock(debug=0)
        config.core.debug = True
        core.adjust_verbosity(mock_ctx_inst)

    @mock.patch('crmsh.report.core.adjust_verbosity')
    @mock.patch('crmsh.report.core.config')
    @mock.patch('json.loads')
    @mock.patch('crmsh.report.core.logger', spec=crmsh.log.DEBUG2Logger)
    def test_load_context(self, mock_logger, mock_json_loads, mock_config, mock_verbosity):
        class Context:
            def __str__(self):
                return "data"
            def __setitem__(self, key, value):
                self.__dict__[key] = value

        sys.argv = ["arg1", "arg2", "arg3"]
        mock_config.report = mock.Mock(verbosity=None)
        mock_json_loads.return_value = {"key": "value", "debug": "true"}
        mock_ctx_inst = Context()
        core.load_context(mock_ctx_inst)
        mock_logger.debug2.assert_called_once_with("Loading context from collector: data")

    @mock.patch('crmsh.report.core.adjust_verbosity')
    @mock.patch('crmsh.report.core.process_arguments')
    @mock.patch('crmsh.utils.check_empty_option_value')
    @mock.patch('crmsh.report.core.add_arguments')
    def test_parse_arguments(self, mock_parse, mock_check_space, mock_process, mock_verbosity):
        mock_args = mock.Mock(option1="value1")
        mock_parse.return_value = mock_args
        mock_ctx_inst = mock.Mock()

        core.parse_arguments(mock_ctx_inst)
        self.assertEqual(mock_ctx_inst.option1, "value1")

        mock_check_space.assert_called_once_with(mock_args)
        mock_process.assert_called_once_with(mock_ctx_inst)

    def test_is_collector(self):
        sys.argv = ["report", "__collector"]
        self.assertEqual(core.is_collector(), True)

    @mock.patch('crmsh.report.core.push_data')
    @mock.patch('crmsh.report.core.collect_logs_and_info')
    @mock.patch('crmsh.report.core.setup_workdir')
    @mock.patch('crmsh.report.core.load_context')
    @mock.patch('crmsh.report.core.is_collector')
    @mock.patch('crmsh.report.core.Context')
    def test_run_impl_collector(self, mock_context, mock_collector, mock_load, mock_setup, mock_collect_info, mock_push):
        mock_context.return_value = mock.Mock()
        mock_ctx_inst = mock_context.return_value
        mock_collector.side_effect = [True, True]

        core.run_impl()

        mock_context.assert_called_once_with()
        mock_collector.assert_has_calls([mock.call(), mock.call()])
        mock_load.assert_called_once_with(mock_ctx_inst)
        mock_setup.assert_called_once_with(mock_ctx_inst)
        mock_collect_info.assert_called_once_with(mock_ctx_inst)
        mock_push.assert_called_once_with(mock_ctx_inst)

    @mock.patch('crmsh.report.core.process_results')
    @mock.patch('crmsh.report.core.collect_for_nodes')
    @mock.patch('crmsh.report.core.load_context_trace_dir_list')
    @mock.patch('crmsh.report.core.find_ssh_user')
    @mock.patch('crmsh.report.core.setup_workdir')
    @mock.patch('crmsh.report.core.load_context_attributes')
    @mock.patch('crmsh.report.core.parse_arguments')
    @mock.patch('crmsh.report.core.is_collector')
    @mock.patch('crmsh.report.core.Context')
    def test_run_impl(self, mock_context, mock_collector, mock_parse, mock_load, mock_setup, mock_find_ssh, mock_load_context_trace_dir_list, mock_collect, mock_process_results):
        mock_context.return_value = mock.Mock()
        mock_ctx_inst = mock_context.return_value
        mock_collector.side_effect = [False, False]

        core.run_impl()

        mock_context.assert_called_once_with()
        mock_collector.assert_has_calls([mock.call(), mock.call()])
        mock_parse.assert_called_once_with(mock_ctx_inst)
        mock_load.assert_called_once_with(mock_ctx_inst)
        mock_setup.assert_called_once_with(mock_ctx_inst)
        mock_find_ssh.assert_called_once_with(mock_ctx_inst)
        mock_load_context_trace_dir_list.assert_called_once_with(mock_ctx_inst)
        mock_collect.assert_called_once_with(mock_ctx_inst)
        mock_process_results.assert_called_once_with(mock_ctx_inst)

    @mock.patch('logging.Logger.error')
    @mock.patch('crmsh.report.core.run_impl')
    def test_run_exception_generic(self, mock_run, mock_log_error):
        mock_run.side_effect = utils.ReportGenericError("error")
        with self.assertRaises(SystemExit) as err:
            core.run()
        mock_log_error.assert_called_once_with("error")

    @mock.patch('crmsh.report.utils.print_traceback')
    @mock.patch('crmsh.report.core.run_impl')
    def test_run_exception(self, mock_run, mock_print):
        mock_run.side_effect = UnicodeDecodeError("encoding", b'', 0, 1, "error")
        with self.assertRaises(SystemExit) as err:
            core.run()
        mock_print.assert_called_once_with()

    @mock.patch('argparse.HelpFormatter')
    @mock.patch('argparse.ArgumentParser')
    def test_add_arguments_help(self, mock_argparse, mock_formatter):
        mock_argparse_inst = mock.Mock()
        mock_argparse.return_value = mock_argparse_inst
        mock_args_inst = mock.Mock(help=True)
        mock_argparse_inst.parse_args.return_value = mock_args_inst

        with self.assertRaises(SystemExit):
            core.add_arguments()

        mock_argparse_inst.print_help.assert_called_once_with()

    @mock.patch('crmsh.report.core.config')
    @mock.patch('argparse.HelpFormatter')
    @mock.patch('argparse.ArgumentParser')
    def test_add_arguments(self, mock_argparse, mock_formatter, mock_config):
        mock_argparse_inst = mock.Mock()
        mock_argparse.return_value = mock_argparse_inst
        mock_args_inst = mock.Mock(help=False, debug=True)
        mock_argparse_inst.parse_args.return_value = mock_args_inst
        mock_config.report = mock.Mock(verbosity=False)

        core.add_arguments()

    @mock.patch('crmsh.report.core.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.to_ascii')
    @mock.patch('crmsh.report.core.ShellUtils')
    def test_push_data(self, mock_sh_utils, mock_to_ascii, mock_logger):
        mock_sh_utils_inst = mock.Mock()
        mock_sh_utils.return_value = mock_sh_utils_inst
        mock_sh_utils_inst.get_stdout_stderr.return_value = (0, "data", "error")
        mock_to_ascii.return_value = "error"
        mock_ctx_inst = mock.Mock(work_dir="/opt/work_dir", main_node="node1", me="node1")

        with self.assertRaises(utils.ReportGenericError) as err:
            core.push_data(mock_ctx_inst)
        self.assertEqual("error", str(err.exception))

        mock_logger.debug2.assert_called_once_with("Pushing data from node1:/opt/work_dir to node1")
        mock_sh_utils_inst.get_stdout_stderr.assert_called_once_with("cd /opt/work_dir/.. && tar -h -c node1", raw=True)

    @mock.patch('crmsh.report.core.finalword')
    @mock.patch('shutil.move')
    @mock.patch('crmsh.report.utils.create_description_template')
    @mock.patch('crmsh.report.utils.analyze')
    def test_process_results_no_compress(self, mock_analyze, mock_create, mock_move, mock_final):
        mock_ctx_inst = mock.Mock(speed_up=True, work_dir="/opt/work_dir", dest_dir="/opt/user", no_compress=True)
        core.process_results(mock_ctx_inst)
        mock_analyze.assert_called_once_with(mock_ctx_inst)
        mock_create.assert_called_once_with(mock_ctx_inst)
        mock_final.assert_called_once_with(mock_ctx_inst)
        mock_move.assert_called_once_with(mock_ctx_inst.work_dir, mock_ctx_inst.dest_dir)

    @mock.patch('crmsh.report.core.finalword')
    @mock.patch('crmsh.sh.cluster_shell')
    @mock.patch('crmsh.report.core.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.report.utils.create_description_template')
    @mock.patch('crmsh.report.utils.analyze')
    @mock.patch('crmsh.report.utils.do_sanitize')
    def test_process_results(self, mock_sanitize, mock_analyze, mock_create, mock_debug2, mock_run, mock_final):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_or_raise_error = mock.Mock()
        mock_ctx_inst = mock.Mock(speed_up=False, work_dir="/opt/work_dir", dest_dir="/opt/user", no_compress=False, dest="report", compress_prog="tar", compress_suffix=".bz2")
        core.process_results(mock_ctx_inst)
        mock_sanitize.assert_called_once_with(mock_ctx_inst)
        mock_analyze.assert_called_once_with(mock_ctx_inst)
        mock_create.assert_called_once_with(mock_ctx_inst)
        mock_final.assert_called_once_with(mock_ctx_inst)

    @mock.patch('crmsh.report.utils.print_traceback')
    @mock.patch('crmsh.report.core.getmembers')
    @mock.patch('multiprocessing.cpu_count')
    @mock.patch('multiprocessing.Pool')
    def test_collect_logs_and_info(self, mock_pool, mock_cpu_count, mock_getmember, mock_print):
        mock_cpu_count.return_value = 4
        mock_pool_inst = mock.Mock()
        mock_pool.return_value = mock_pool_inst
        mock_pool_inst.apply_async = mock.Mock()
        mock_async_inst1 = mock.Mock()
        mock_async_inst2 = mock.Mock()
        mock_pool_inst.apply_async.side_effect = [mock_async_inst1, mock_async_inst2]
        mock_async_inst1.get = mock.Mock()
        mock_async_inst2.get = mock.Mock(side_effect=ValueError)
        mock_pool_inst.close = mock.Mock()
        mock_pool_inst.join = mock.Mock()
        mock_getmember.return_value = [("collect_func1", None), ("collect_func2", None)]
        collect.collect_func1 = mock.Mock()
        collect.collect_func2 = mock.Mock()
        mock_ctx_inst = mock.Mock()

        core.collect_logs_and_info(mock_ctx_inst)
        mock_pool.assert_called_once_with(3)

    @mock.patch('multiprocessing.Process')
    @mock.patch('logging.Logger.info')
    @mock.patch('crmsh.report.core.start_collector')
    def test_collect_for_nodes(self, mock_start_collector, mock_info, mock_process):
        mock_ctx_inst = mock.Mock(
            node_list=["node1", "node2"],
            passwordless_shell_for_nodes={"node1": mock.Mock(crmsh.report.sh.Shell)},
            ssh_user=""
        )
        mock_process_inst = mock.Mock()
        mock_process.return_value = mock_process_inst
        core.collect_for_nodes(mock_ctx_inst)

    def test_process_arguments_value_error(self):
        mock_ctx_inst = mock.Mock(from_time=123, to_time=100)
        with self.assertRaises(ValueError) as err:
            core.process_arguments(mock_ctx_inst)
            self.assertEqual("The start time must be before the finish time", str(err.exception))

    @mock.patch('crmsh.report.core.local_mode')
    @mock.patch('crmsh.utils.list_cluster_nodes')
    def test_process_node_list_exception(self, mock_list_nodes, mock_local_mode):
        mock_local_mode.return_value = False
        config.core.no_ssh = False
        mock_ctx_inst = mock.Mock(node_list=[], single=False)
        mock_list_nodes.return_value = []
        with self.assertRaises(utils.ReportGenericError) as err:
            core.process_node_list(mock_ctx_inst)
            self.assertEqual("Could not figure out a list of nodes; is this a cluster node?", str(err.exception))

    @mock.patch('crmsh.utils.get_reachable_node_list')
    @mock.patch('crmsh.report.core.local_mode')
    @mock.patch('crmsh.utils.list_cluster_nodes')
    def test_process_node_list_single(self, mock_list_nodes, mock_local_mode, mock_reachable):
        mock_local_mode.return_value = True
        mock_ctx_inst = mock.Mock(node_list=["node1", "node2"], single=True, me="node1")
        core.process_node_list(mock_ctx_inst)

    @mock.patch('crmsh.utils.get_reachable_node_list')
    @mock.patch('crmsh.report.core.local_mode')
    @mock.patch('logging.Logger.error')
    @mock.patch('crmsh.utils.list_cluster_nodes')
    def test_process_node_list(self, mock_list_nodes, mock_error, mock_local_mode, mock_reachable):
        mock_local_mode.return_value = False
        mock_ctx_inst = mock.Mock(node_list=["node1", "node2"], single=False, me="node1")
        mock_reachable.return_value = ["node1"]
        core.process_node_list(mock_ctx_inst)
        self.assertEqual(mock_ctx_inst.node_list, ["node1"])

    @mock.patch('crmsh.report.core.process_node_list')
    @mock.patch('crmsh.report.core.process_dest')
    def test_process_arguments(self, mock_dest, mock_node_list):
        mock_ctx_inst = mock.Mock(from_time=123, to_time=150)
        core.process_arguments(mock_ctx_inst)


    @mock.patch('subprocess.Popen')
    @mock.patch('ast.literal_eval')
    def test_start_collector(self, mock_literal_eval, mock_popen):
        mock_shell = mock.Mock(crmsh.report.sh.Shell)
        mock_context = mock.Mock(
            ssh_user=None,
            passwordless_shell_for_nodes={'node1': mock_shell},
            __str__=mock.Mock(return_value='{"foo":"bar"}')
        )
        mock_shell.subprocess_run_without_input.return_value = mock.Mock(
            returncode=0,
            stdout=b'',
            stderr=b'',
        )
        core.start_collector('node1', mock_context)
        mock_shell.subprocess_run_without_input.assert_called_once_with(
            """/usr/sbin/crm report __collector '{"foo":"bar"}'""",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    @mock.patch('crmsh.sh.ShellUtils.get_stdout')
    @mock.patch('ast.literal_eval')
    @mock.patch('logging.Logger.warning')
    def test_start_collector_warn(self, mock_warning, mock_literal_eval, mock_get_stdout):
        mock_shell = mock.Mock(crmsh.report.sh.Shell)
        mock_context = mock.Mock(
            ssh_user=None,
            passwordless_shell_for_nodes={'node1': mock_shell},
            __str__=mock.Mock(return_value='{"foo":"bar"}')
        )
        mock_shell.subprocess_run_without_input.return_value = mock.Mock(
            returncode=1,
            stdout=b'',
            stderr=b'asdfgh',
        )
        core.start_collector('node1', mock_context)
        mock_shell.subprocess_run_without_input.assert_called_once_with(
            """/usr/sbin/crm report __collector '{"foo":"bar"}'""",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        mock_warning.assert_called_with(
            'Failed to run collector on %s: %s: %s',
            'node1', 1, 'asdfgh',
        )
        mock_literal_eval.assert_not_called()
        mock_get_stdout.assert_not_called()
